#! /usr/bin/env python

#######################################
# A script to calculate statistics from
# a text file using NumPy
# Author: <YOUR NAME>
# Email: <YOUR EMAIL>
# Date: DD/MM/YYYY
# Version: 1.0
#######################################

import numpy
import scipy
# Import scipy stats functions we need
import scipy.stats as spstats
# Import plotting library as plt
import matplotlib.pyplot as plt

class CalculateStatistics (object):
    
    def run(self):
        # Set up lists to hold input diameters and heights
        # A seperate list is used for each species
        beechDiameter = list()
        beechHeight = list()
        ashDiameter = list()
        ashHeight = list()
        birchDiameter = list()
        birchHeight = list()
        oakDiameter = list()
        oakHeight = list()
        sycamoreDiameter = list()
        sycamoreHeight = list()
        otherDiameter = list()
        otherHeight = list()
        
        # Open input and output files
        inFileName = 'PenglaisWoodsData.csv'
        outFileName = 'PenglaisWoodsStats.csv'
        inFile = open(inFileName, 'r')
        outFile = open(outFileName,'w')
        
        # Iterate through the input file and save diameter and height
        # into lists, based on species
        header = True 
        for eachLine in inFile:
            if header: # Skip header row
                print('Skipping header row')
                header = False
            else:
                substrs = eachLine.split(',',eachLine.count(','))
                
                species = substrs[3]
                if substrs[4].isdigit: # Check diameter is a number 
                    diameter = float(substrs[4])
                    height = float(substrs[10])
                
                    if species == 'BEECH':
                        beechDiameter.append(diameter)
                        beechHeight.append(height)
                    elif species == 'ASH':
                        ashDiameter.append(diameter)
                        ashHeight.append(height)
                    elif species == 'BIRCH':
                        birchDiameter.append(diameter)
                        birchHeight.append(height)
                    elif species == 'OAK':
                        oakDiameter.append(diameter)
                        oakHeight.append(height)
                    elif species == 'SYC':
                        sycamoreDiameter.append(diameter)
                        sycamoreHeight.append(height)
                    else:
                        otherDiameter.append(diameter)
                        otherHeight.append(height)
        
        # Convert to NumPy arrays
        beechDiameter = numpy.array(beechDiameter)
        ashDiameter = numpy.array(ashDiameter)
        birchDiameter = numpy.array(birchDiameter)
        oakDiameter = numpy.array(oakDiameter)
        sycamoreDiameter = numpy.array(sycamoreDiameter)
        otherDiameter = numpy.array(otherDiameter)
        
        beechHeight = numpy.array(beechHeight)
        ashHeight = numpy.array(ashHeight)
        birchHeight = numpy.array(birchHeight)
        oakHeight = numpy.array(oakHeight)
        sycamoreHeight = numpy.array(sycamoreHeight)
        otherHeight = numpy.array(otherHeight)
        
        # Write header line to output file
        headerLine = 'species, meanDiameter, medianDiameter, stDevDiameter\n'
        outFile.write(headerLine)
    
        # Calculate statistics for each species and write to file
        outLine = 'Beech,' + self.createStatsLine(beechDiameter) + '\n'
        outFile.write(outLine)
        outLine = 'Ash,' + self.createStatsLine(ashDiameter) + '\n'
        outFile.write(outLine)
        outLine = 'Birch,' + self.createStatsLine(birchDiameter) + '\n'
        outFile.write(outLine)
        outLine = 'Oak,' + self.createStatsLine(oakDiameter) + '\n'
        outFile.write(outLine)
        outLine = 'Sycamore,' + self.createStatsLine(sycamoreDiameter) + '\n'
        outFile.write(outLine)
        outLine = 'Other,' + self.createStatsLine(otherDiameter) + '\n'
        outFile.write(outLine)
        
        print('Statistics written to: ' + outFileName)
        
        # Fit line to each file and save out plot
        
        # Set output directory for plots
        outDIR = './' 

        # Plot linear regression for Beech
        print('Generating plot:')
        self.plotLinearRegression(beechDiameter, beechHeight, outDIR + 'beech.pdf')
        
    def plotLinearRegression(self, inDiameterArray, inHeightArray, outPlotName):
        # Perform fit
        (aCoeff,bCoeff,rVal,pVal,stdError) = spstats.linregress(numpy.log(inDiameterArray), inHeightArray)
        
        # Use fits to predict height for a range of diameters
        testDiameter = numpy.arange(min(inDiameterArray), max(inDiameterArray), 1)
        predictHeight = (aCoeff * numpy.log(testDiameter)) + bCoeff
        
        # Create a string, showing the form of the equation (with fitted coefficients) 
        # and r squared value
        # Coefficients are rounded to two decimal places.        
        equation = str(round(aCoeff,2)) + 'log(D) ' + str(round(bCoeff,2)) + \
        ' (r$^2$ = ' + str(round(rVal**2,2)) + ')'
        
        # Plot fit against origional data
        plt.plot(inDiameterArray, inHeightArray,'.')
        plt.plot(testDiameter, predictHeight)
        plt.xlabel('Diameter (cm)')
        plt.ylabel('Height (m)')
        plt.legend(['measured data',equation])
        
        # Save plot
        plt.savefig(outPlotName, dpi=200, format='PDF')
        
                
    def createStatsLine(self, inArray):
        # Calculate statsistics for array and return output line.
        meanArray = numpy.mean(inArray)
        medianArray = numpy.median(inArray)
        stDevArray = numpy.std(inArray)
        
        # Create output line with stats
        statsLine = str(meanArray) + ',' + str(medianArray) + ',' + str(stDevArray) 
        return statsLine
    
    def calcBiomassStatsLine(self, inDiameterArray, inHeightArray, inSpecies):
        # Calculates biomass, calculates stats from biomass and returns output line
        biomassStatsLine = ''
        try:
            # Calculate biomass
            biomass = self.calcBiomass(inDiameterArray, inHeightArray, inSpecies)
            # Calculate stats from biomass
            meanBiomass = numpy.mean(biomass)
            medianBiomass = numpy.median(biomass)
            stDevBiomass = numpy.std(biomass)
            
            # Create output line
            biomassStatsLine = str(meanBiomass) + ',' + str(medianBiomass) + ','\
            + str(stDevBiomass) 
            
        except Exception:
            # Catch exception and write 'na' for all values
            biomassStatsLine = 'na,na,na'
        
        return biomassStatsLine
    
    def calcBiomass(self, inDiameterArray, inHeightArray, inSpecies):        
        if inSpecies == 'BEECH':
            a = 0.014306
            b = 0.0000748
            specificGravity = 0.56
        else: # Raise exception is species is not recognised
            raise Exception('Species not recognised')
            
        # Calcualte volume
        volume = a + ((b*(inDiameterArray)**2) * (inHeightArray**0.75))
        # Calculate biomass
        biomass = volume * specificGravity
        # Return biomass
        return biomass
        
if __name__ == '__main__':
    obj = CalculateStatistics()
    obj.run()
