#!/usr/bin/env python

# Import the system library
import sys
# Import the rat modele from RIOS
from rios import rat
# Import the numpy library
import numpy
# Import the gdal library
import osgeo.gdal as gdal
# Import the python Argument parser
import argparse

# Thresholds which have been globally defined 
# so they can easily be changed and therefore
# change everywhere they are used.
URBAN_MASK_THRES = 0.05
CULT_MASK_THRES = 0.05
WBI_PRE_THRES = 1
WBI_PEAK_THRES = 1
WBI_POST_THRES = 1
FDI_PRE_THRES = -100
FDI_PEAK_THRES = -40
FDI_POST_THRES = -115
PSRI_PRE_THRES = -0.2
REP_PEAK_THRES = -0.005
WOODY_PRE_THRES_BG = 0.09
WOODY_PEAK_THRES_CG = 0.195

# A function for classifying the first part of level 1
def classifyLevel1FromImg(urbanMask, wbiPeak, fdiPeak, wbiPost, 
                          fdiPost, wbiPre, fdiPre, psriPre, repPeak):
    # Create Output Array
    l1P1 = numpy.empty_like(urbanMask, dtype=numpy.dtype('a255'))
    l1P1[...] = "NA"
    # Urban
    l1P1 = numpy.where(numpy.logical_and(l1P1 == "NA", urbanMask > URBAN_MASK_THRES), 
                       "Urban", l1P1)
    # Water
    l1P1 = numpy.where(numpy.logical_and(l1P1 == "NA", 
                       numpy.logical_or(wbiPre >= WBI_PRE_THRES, 
                                        wbiPeak >= WBI_PEAK_THRES)), 
                       "Water", l1P1)
    # Photosynthetic Vegetation 
    l1P1 = numpy.where(numpy.logical_and(l1P1 == "NA", 
                       numpy.logical_or(fdiPeak > FDI_PEAK_THRES, 
                                        fdiPost > FDI_POST_THRES)), 
                       "Photosynthetic Vegetated", l1P1)    
    # Non PhotoSynthetic Vegetation
    l1P1 = numpy.where(numpy.logical_and(l1P1 == "NA", 
                                         psriPre >= PSRI_PRE_THRES), 
                    "Non Photosynthetic Vegetated", l1P1)
    # Non Submerged Aquatic Veg
    l1P1 = numpy.where(numpy.logical_and(l1P1 == "NA", 
                       numpy.logical_and(repPeak >= REP_PEAK_THRES, 
                                         wbiPost <= WBI_POST_THRES)), 
                       "Non Submerged Aquatic Vegetated", l1P1)
    return l1P1

# A function for classifying the second part of level 1
def classifyLevel1Assign(classLevel1Img):
    # Create Output Array
    level1 = numpy.empty_like(classLevel1Img, dtype=numpy.dtype('a255'))
    level1[...] = "NA"
    # Non Vegetated
    level1 = numpy.where(numpy.logical_or(classLevel1Img == "NA", 
                         numpy.logical_or(classLevel1Img == "Water", 
                                          classLevel1Img == "Urban")), 
                         "Non Vegetated", level1)
    # Vegetated
    level1 = numpy.where(numpy.logical_or(
                         classLevel1Img == "Photosynthetic Vegetated", 
                         classLevel1Img == "Non Photosynthetic Vegetated", 
                         classLevel1Img == "Non Submerged Aquatic Vegetated"), 
                         "Vegetated", level1)
    return level1

# A function for classifying level 2
def classifyLevel2(wbiPre, wbiPeak, wbiPost, classLevel1Img):
    # Create Output Array
    level2 = numpy.empty_like(classLevel1Img, dtype=numpy.dtype('a255'))
    level2[...] = "NA"
    
    # Terrestrial Non Vegetated
    level2 = numpy.where(numpy.logical_or(classLevel1Img == "NA", 
                                          classLevel1Img == "Urban"), 
                         "Terrestrial Non Vegetated", level2)
    # Aquatic Non Vegetated
    level2 = numpy.where(numpy.logical_and(
                         numpy.logical_not(classLevel1Img == "Urban"), 
                         numpy.logical_or(wbiPre > 1, wbiPeak > 1)), 
                         "Aquatic Non Vegetated", level2)
    # Terrestrial Vegetated
    level2 = numpy.where(numpy.logical_or(classLevel1Img == "Photosynthetic Vegetated", 
                         classLevel1Img == "Non Photosynthetic Vegetated"), 
                         "Terrestrial Vegetated", level2)
    # Aquatic Vegetated
    level2 = numpy.where(classLevel1Img == "Non Submerged Aquatic Vegetated", 
                         "Aquatic Vegetated", level2)
    return level2

# A function for classifying level 3
def classifyLevel3(classLevel2, cult, urban):
    # Create Output Array
    level3 = numpy.empty_like(classLevel2, dtype=numpy.dtype('a255'))
    level3[...] = "NA"
    # Cultivated Terrestrial Vegetated
    level3 = numpy.where(numpy.logical_and(
                         classLevel2 == "Terrestrial Vegetated", cult > CULT_MASK_THRES), 
                         "Cultivated Terrestrial Vegetated", level3)
    # Natural Terrestrial Vegetated
    level3 = numpy.where(numpy.logical_and(numpy.logical_not
                        (level3 == "Cultivated Terrestrial Vegetated"), 
                        classLevel2 == "Terrestrial Vegetated"), 
                        "Natural Terrestrial Vegetated", level3)
    # Cultivated Aquatic Vegetated
    level3 = numpy.where(numpy.logical_and(classLevel2 == "Aquatic Vegetated", 
                         cult > CULT_MASK_THRES), "Cultivated Aquatic Vegetated", level3)
    # Natural Aquatic Vegetated
    level3 = numpy.where(numpy.logical_and(numpy.logical_not
                        (level3 == "Cultivated Aquatic Vegetated"), 
                        classLevel2 == "Aquatic Vegetated"), 
                        "Natural Aquatic Vegetated", level3)
    # Artificial Surface
    level3 = numpy.where(numpy.logical_and(classLevel2 == "Terrestrial Non Vegetated", 
                         urban > URBAN_MASK_THRES), "Artificial Surface", level3)
    # Natural Surface
    level3 = numpy.where(numpy.logical_and(numpy.logical_not
                        (level3 == "Artificial Surface"), 
                        classLevel2 == "Terrestrial Non Vegetated"), 
                        "Natural Surface", level3)
    # Natural Water
    level3 = numpy.where(classLevel2 == "Aquatic Non Vegetated", 
                         "Natural Water", level3)
    return level3

def runClassification(fname):
    # Open the GDAL Dataset so it is just opened once
    # and reused rather than each rios call reopening
    # the image file which will large attribute tables
    # can be slow.
    ratDataset = gdal.Open( fname, gdal.GA_Update )
    # Check the image file was openned correctly.
    if not ratDataset == None:
        # Provide feedback to the user.
        print("Import Columns.")
        urban = rat.readColumn(ratDataset, "PropUrban")
        cult = rat.readColumn(ratDataset, "PropCult")
        
        # Read in the RAT columns for the Pre-Flush image
        PreCoastal = rat.readColumn(ratDataset, "MarB1")
        PreBlue = rat.readColumn(ratDataset, "MarB2")
        PreRed = rat.readColumn(ratDataset, "MarB5")
        PreRedEdge = rat.readColumn(ratDataset, "MarB6")
        PreNIR1 = rat.readColumn(ratDataset, "MarB7")
    
        # Read in the RAT columns for the Peak-flush image.
        PeakCoastal = rat.readColumn(ratDataset, "JulyB1")
        PeakBlue = rat.readColumn(ratDataset, "JulyB2")
        PeakRed = rat.readColumn(ratDataset, "JulyB5")
        PeakRedEdge = rat.readColumn(ratDataset, "JulyB6")
        PeakNIR1 = rat.readColumn(ratDataset, "JulyB7")
        PeakNIR2 = rat.readColumn(ratDataset, "JulyB8")
    
        # Read in the RAT columns for the Post-flush image.
        PostCoastal = rat.readColumn(ratDataset, "NovB1")
        PostBlue = rat.readColumn(ratDataset, "NovB2")
        PostRedEdge = rat.readColumn(ratDataset, "NovB6")
        PostNIR1 = rat.readColumn(ratDataset, "NovB7")
    
        # Provide more feedback to the user. 
        print("Calculate Indices.")
        # As all the columns are numpy arrays then
        # we can do numpy arithmatic between the 
        # arrays to calculate new arrays, such as 
        # indices.
        wbiPre = PreBlue/PreNIR1
        fdiPre = PreNIR1 - (PreRedEdge + PreCoastal)
        psriPre = (PreRed - PreBlue)/PreRedEdge
    
        wbiPeak = PeakBlue/PeakNIR1
        fdiPeak = PeakNIR1 - (PeakRedEdge + PeakCoastal)
        repPeak = PeakRedEdge - (PeakNIR2 - PeakRed)
    
        wbiPost = PostBlue/PostNIR1
        fdiPost = PostNIR1 - (PostRedEdge + PostCoastal)
        
        # Call the function which classifies the first part 
		# of the level 1 classification
        print("Classifying Level 1")
        classLevel1Img = classifyLevel1FromImg(urban, 
                                               wbiPeak, 
                                               fdiPeak, 
                                               wbiPost, 
                                               fdiPost, 
                                               wbiPre, 
                                               fdiPre, 
                                               psriPre, 
                                               repPeak)
        # Write the first part of the level 1 classification 
        # back into the input file.
        rat.writeColumn(ratDataset, "ClassLevel1Part1", classLevel1Img)
        # Call function a produce the level 1 classification
        classLevel1 = classifyLevel1Assign(classLevel1Img)
        # Write the level 1 classification to the image
        rat.writeColumn(ratDataset, "ClassLevel1", classLevel1)
    
        # Call the function which classifies the level 2 of the classification.
        print("Classifying Level 2")
        classLevel2 = classifyLevel2(wbiPre, wbiPeak, wbiPost, classLevel1Img)
        # Write the level 2 classification to the image.
        rat.writeColumn(ratDataset, "ClassLevel2", classLevel2)
    
        # Call the function which classifies level 3 of the classification
        print("Classifying Level 3")
        classLevel3 = classifyLevel3(classLevel2, cult, urban)
        #  Write the level 3 classification to the image.
        rat.writeColumn(ratDataset, "ClassLevel3", classLevel3)
    else:
        print("Image could not be openned")

# This is the first part of the script to 
# be executed.
if __name__ == '__main__':
    # Create the command line options 
    # parser.
    parser = argparse.ArgumentParser()
    # Define the argument for specifying the input file.
    parser.add_argument("-i", "--input", type=str, 
                        help="Specify the input image file.")
    # Call the parser to parse the arguments.
    args = parser.parse_args()
    
    # Check that the input parameter has been specified.
    if args.input == None:
        # Print an error message if not and exit.
        print("Error: No input image file provided.")
        sys.exit()
        
    # Run the classification
    runClassification(args.input)
